/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.executable;

import static org.fortiss.af3.component.utils.ComponentArchitectureUtils.findSourcePort;

import java.util.HashMap;

import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.deployment.generator.TransmissionCatalog;
import org.fortiss.af3.deployment.model.ComponentAllocation;
import org.fortiss.af3.deployment.model.Deployment;
import org.fortiss.af3.platform.raspberry.model.RaspberryPi;

/**
 * Can Transmission Catalog
 * 
 * @author eder
 * @author hoelzl
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating RED Hash:
 */
public class CanTransmissionCatalog extends TransmissionCatalog {
	/** Mapping from atomic component to deployed execution unit. */
	private HashMap<Component, RaspberryPi> componentToRaspberry = new HashMap<>();
	/** Mapping from execution unit to CAN IDs. */
	private HashMap<RaspberryPi, Integer> raspberryCanId = new HashMap<>();
	/** Counter for raspberry CAN IDs. */
	private int canIdCounter = 0xF;
	/** Mapping from atomic input port to atomic output port. */
	private HashMap<InputPort, OutputPort> atomicInputToOutput = new HashMap<>();

	/** Constructor. */
	public CanTransmissionCatalog(Deployment deployment) {
		for(ComponentAllocation ca : deployment.getComponentAllocations()) {
			RaspberryPi raspi = (RaspberryPi)ca.getExecutionUnit();
			if(ca.getComponent().getSubComponents().isEmpty()) {
				analyzeAtomic(ca.getComponent(), raspi);
				continue;
			}
			for(Component atomic : ca.getComponent().findAtomicComponents()) {
				analyzeAtomic(atomic, raspi);
			}
		}
	}

	/** Analyzes the given component. */
	private void analyzeAtomic(Component component, RaspberryPi raspi) {
		componentToRaspberry.put(component, raspi);
		if(!raspberryCanId.containsKey(raspi)) {
			raspberryCanId.put(raspi, canIdCounter++);
		}
		for(InputPort inport : component.getInputPorts()) {
			Port sourcePort = findSourcePort(null, inport);
			if(sourcePort instanceof OutputPort) {
				atomicInputToOutput.put(inport, (OutputPort)sourcePort);
			}
		}
	}

	/** Returns whether this input port is connected to a remote source. */
	public boolean hasRemoteSource(InputPort inport) {
		OutputPort outport = atomicInputToOutput.get(inport);
		if(outport == null) {
			return false;
		}
		RaspberryPi inPi = componentToRaspberry.get(inport.getComponent());
		RaspberryPi outPi = componentToRaspberry.get(outport.getComponent());
		return inPi != outPi;
	}

	/** Returns whether the given output port is needed remotely. */
	public boolean isOutputNeededRemotely(OutputPort outport) {
		return atomicInputToOutput.containsValue(outport);
	}

	/** Returns the atomic (remote) source port for this input port. */
	public OutputPort getSourcePort(InputPort inport) {
		return atomicInputToOutput.get(inport);
	}

	/** Returns the {@link RaspberryPi} for the given port. */
	public RaspberryPi getRaspberryPi(Port port) {
		return getRaspberryPi(port.getComponent());
	}

	/** Returns the {@link RaspberryPi} for the given component. */
	public RaspberryPi getRaspberryPi(Component component) {
		return componentToRaspberry.get(component);
	}

	/** Returns the CAN id for the given {@link RaspberryPi}. */
	public int getCanId(RaspberryPi raspi) {
		return raspberryCanId.get(raspi);
	}
}
