/*--------------------------------------------------------------------------+
$Id: MainGenerator.java 23939 2018-04-10 12:20:41Z eder $
|                                                                          |
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.executable;

import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getSingleUnitMainCFile;

import java.util.HashMap;
import java.util.List;

import org.conqat.lib.commons.collections.Pair;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.generator.common.model.source.AbstractUnit;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.PlatformConnectorUnit;
import org.fortiss.af3.platform.model.Receiver;
import org.fortiss.af3.platform.model.Transmitter;
import org.fortiss.af3.platform.raspberry.model.ActuatorPWM;
import org.fortiss.af3.platform.raspberry.model.ConsoleOutput;
import org.fortiss.af3.platform.raspberry.model.RaspberryPi;
import org.fortiss.af3.platform.raspberry.model.brick.AccelerationXSensor;
import org.fortiss.af3.platform.raspberry.model.brick.AccelerationYSensor;
import org.fortiss.af3.platform.raspberry.model.brick.AccelerationZSensor;
import org.fortiss.af3.platform.raspberry.model.brick.ActuatorDigits;
import org.fortiss.af3.platform.raspberry.model.brick.AngularVelocityXSensor;
import org.fortiss.af3.platform.raspberry.model.brick.AngularVelocityYSensor;
import org.fortiss.af3.platform.raspberry.model.brick.AngularVelocityZSensor;
import org.fortiss.af3.platform.raspberry.model.brick.LEDButtonBActuator;
import org.fortiss.af3.platform.raspberry.model.brick.LEDButtonGActuator;
import org.fortiss.af3.platform.raspberry.model.brick.LEDButtonRActuator;
import org.fortiss.af3.platform.raspberry.model.brick.LaserRangeSensor;
import org.fortiss.af3.platform.raspberry.model.brick.UIDUnit;
import org.fortiss.af3.platform.raspberry.model.brick.UltraSonicSensor;
import org.fortiss.af3.platform.raspberry.model.camera.CameraDetectionStateLeft;
import org.fortiss.af3.platform.raspberry.model.camera.CameraDetectionStateRight;
import org.fortiss.af3.platform.raspberry.model.camera.CameraDistanceLeft;
import org.fortiss.af3.platform.raspberry.model.camera.CameraDistanceRight;
import org.fortiss.af3.platform.raspberry.model.camera.CameraYawAngle;
import org.fortiss.af3.platform.raspberry.model.gamepad.Button1;
import org.fortiss.af3.platform.raspberry.model.gamepad.Button2;
import org.fortiss.af3.platform.raspberry.model.gamepad.Button3;
import org.fortiss.af3.platform.raspberry.model.gamepad.Button4;
import org.fortiss.af3.platform.raspberry.model.gamepad.ButtonL2;
import org.fortiss.af3.platform.raspberry.model.gamepad.ButtonR2;
import org.fortiss.af3.platform.raspberry.model.gamepad.GamepadReceiverBase;
import org.fortiss.af3.platform.raspberry.model.motorcontrol.MotorControlInput;
import org.fortiss.af3.platform.raspberry.model.motorcontrol.MotorControlOutput;
import org.fortiss.af3.platform.raspberry.model.rumblepad.ButtonA;
import org.fortiss.af3.platform.raspberry.model.rumblepad.ButtonB;
import org.fortiss.af3.platform.raspberry.model.rumblepad.ButtonHome;
import org.fortiss.af3.platform.raspberry.model.rumblepad.ButtonSelect;
import org.fortiss.af3.platform.raspberry.model.rumblepad.ButtonStart;
import org.fortiss.af3.platform.raspberry.model.rumblepad.ButtonX;
import org.fortiss.af3.platform.raspberry.model.rumblepad.ButtonY;
import org.fortiss.af3.platform.raspberry.model.rumblepad.DPadDown;
import org.fortiss.af3.platform.raspberry.model.rumblepad.DPadLeft;
import org.fortiss.af3.platform.raspberry.model.rumblepad.DPadRight;
import org.fortiss.af3.platform.raspberry.model.rumblepad.DPadUp;
import org.fortiss.af3.platform.raspberry.model.rumblepad.L2_Position;
import org.fortiss.af3.platform.raspberry.model.rumblepad.R2_Position;
import org.fortiss.af3.platform.raspberry.model.rumblepad.RumbleMagnitudeStrong;
import org.fortiss.af3.platform.raspberry.model.rumblepad.RumbleMagnitudeWeak;
import org.fortiss.af3.platform.raspberry.model.rumblepad.RumblepadReceiverBase;
import org.fortiss.tooling.kernel.extension.data.ITransformationContext;

/**
 * Separate class for generating the main.c file.
 * 
 * @author hoelzl
 * @author $Author: eder $
 * @version $Rev: 23939 $
 * @ConQAT.Rating RED Hash:
 */
class MainGenerator {
	private static final int WAITING_SLEEP_IN_MICROS = 250;
	private RaspberryPi executionUnit;
	private List<Pair<ExecutionUnit, Component>> deployedComponents;
	private List<Pair<PlatformConnectorUnit, Port>> deployedPorts;
	private ITransformationContext context;

	/** LED_Button UIDs to RGB value array. */
	private HashMap<String, String[]> ledButtons = new HashMap<>();

	/** LED_Button UIDs to RGB local variable array. */
	private HashMap<String, String[]> ledButtonsLocalVariable = new HashMap<>();

	private boolean useCamera;
	private boolean useGamepad;
	private boolean useRumblepad;
	private boolean useDigits;
	private boolean useLEDButton;
	private boolean usePWM;
	private boolean useUS;
	private boolean useLaser;
	private boolean useACC;
	private boolean useConsole;
	private boolean useRumble;
	private OutputPort rumbleWeakPort = null;
	private OutputPort rumbleStrongPort = null;
	private int nrOfUsSensors = 0;
	private boolean useVesc;

	/** Constructor. */
	public MainGenerator(RaspberryPi executionUnit,
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts, ITransformationContext context) {
		this.executionUnit = executionUnit;
		this.deployedComponents = deployedComponents;
		this.deployedPorts = deployedPorts;
		this.context = context;
		for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
			if(!useGamepad && (p.getFirst() instanceof GamepadReceiverBase)) {
				useGamepad = true;
			}
			if(!useRumblepad && p.getFirst() instanceof RumblepadReceiverBase ||
					p.getFirst() instanceof RumbleMagnitudeWeak ||
					p.getFirst() instanceof RumbleMagnitudeStrong) {
				useRumblepad = true;
			}
			if(!usePWM && p.getFirst() instanceof ActuatorPWM) {
				usePWM = true;
			}
			if(!useUS && p.getFirst() instanceof UltraSonicSensor) {
				useUS = true;
			}
			if(!useLaser && p.getFirst() instanceof LaserRangeSensor) {
				useLaser = true;
			}
			if(!useDigits && p.getFirst() instanceof ActuatorDigits) {
				useDigits = true;
			}
			if((p.getFirst() instanceof LEDButtonRActuator ||
					p.getFirst() instanceof LEDButtonGActuator || p.getFirst() instanceof LEDButtonBActuator)) {
				extractRGBforLEDButton(p);
				useLEDButton = true;

			}
			if(!useACC &&
					(p.getFirst() instanceof AccelerationXSensor ||
							p.getFirst() instanceof AccelerationYSensor ||
							p.getFirst() instanceof AccelerationZSensor ||
							p.getFirst() instanceof AngularVelocityXSensor ||
							p.getFirst() instanceof AngularVelocityYSensor || p.getFirst() instanceof AngularVelocityZSensor)) {
				useACC = true;
			}
			if(!useConsole && p instanceof ConsoleOutput) {
				useConsole = true;
			}

			if(!useVesc &&
					(p.getFirst() instanceof MotorControlOutput || p.getFirst() instanceof MotorControlInput)) {
				useVesc = true;
			}

			if(!useCamera &&
					(p.getFirst() instanceof CameraYawAngle ||
							p.getFirst() instanceof CameraDistanceLeft ||
							p.getFirst() instanceof CameraDistanceRight || p.getFirst() instanceof CameraDetectionStateLeft) ||
					p.getFirst() instanceof CameraDetectionStateRight) {
				useCamera = true;
			}
		}
	}

	/** Creates the main.c file for deployments with a single execution units. */
	public AbstractUnit createSingleUnitMain() {
		String includes = createIncludes(deployedComponents, deployedPorts);
		String initCode = createInitCode(deployedComponents);
		String sensorVariables = createVariables(deployedComponents, deployedPorts);
		String workerCode = createWorkerCode(deployedComponents, deployedPorts);
		return getSingleUnitMainCFile(executionUnit.getName(), executionUnit.getCycleTime(),
				includes, initCode, workerCode, sensorVariables);
	}

	/** Creates the includes of the system headers. */
	private String createIncludes(List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts) {
		StringBuilder sb = new StringBuilder();
		for(Pair<ExecutionUnit, Component> p : deployedComponents) {
			Component c = p.getSecond();
			sb.append("#include <" + c.getName() + "_ID_" + c.getId() + ".h>\n");
		}
		if(useCamera) {
			sb.append("#include <camera_client.h>\n");
		}
		if(useGamepad) {
			sb.append("#include <gamepad.h>\n");
		}
		if(useRumblepad) {
			sb.append("#include <rumblepad.h>\n");
		}
		if(usePWM) {
			sb.append("#include <temp_actuator.h>\n");
		}
		if(useUS || useLaser || useDigits || useACC || useLEDButton) {
			sb.append("#include <ip_connection.h>\n");
		}
		if(useUS) {
			sb.append("#include <bricklet_distance_us.h>\n");
		}
		if(useLaser) {
			sb.append("#include <bricklet_laser_range_finder.h>\n");
		}
		if(useACC) {
			sb.append("#include <brick_imu_v2.h>\n");
		}
		if(useDigits) {
			sb.append("#include <bricklet_segment_display_4x7.h>\n");
		}
		if(useLEDButton) {
			sb.append("#include <bricklet_rgb_led_button.h>\n");
		}
		if(useVesc) {
			sb.append("#include <commands.h>\n");
			sb.append("#include <hal.h>\n");
		}
		sb.append("#include <stdio.h>\n");
		return sb.toString();
	}

	/** Create the worker code. */
	private String createWorkerCode(List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts) {
		StringBuilder sb = new StringBuilder();
		if(useUS || useLaser || useACC) {
			// Get Current time for the sensors
			sb.append("uint64_t curr_time = time_util_get_current_micros();\n");
		}
		for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
			if(p.getFirst() instanceof Receiver && p.getSecond() instanceof InputPort) {
				sb.append(createReadCode((Receiver)p.getFirst(), (InputPort)p.getSecond()));
			}
		}
		for(Pair<ExecutionUnit, Component> p : deployedComponents) {
			Component c = p.getSecond();
			sb.append(makeCall("perform_step", c));
		}
		for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
			if(p.getFirst() instanceof Transmitter && p.getSecond() instanceof OutputPort) {
				sb.append(createWriteCode((Transmitter)p.getFirst(), (OutputPort)p.getSecond()));
			}
		}
		if(useRumble) {
			sb.append(createRumbleCode());
		}
		return sb.toString();
	}

	/** Create the Sensor Variables and functions. */
	private String createVariables(List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts) {
		StringBuilder sb = new StringBuilder();
		if(usePWM) {
			sb.append("int maestro_fd = -1;\n");
		}
		if(useDigits) {
			sb.append("SegmentDisplay4x7 segment_display;\n");
			sb.append("static void set_led_display (SegmentDisplay4x7 *segment_display, uint16_t value, bool showHex) {\n");
			sb.append("static const uint8_t digits[] = {0x3f,0x06,0x5b,0x4f,\n");
			sb.append("                                 0x66,0x6d,0x7d,0x07,\n");
			sb.append("                                 0x7f,0x6f,0x77,0x7c,\n");
			sb.append("                                 0x39,0x5e,0x79,0x71};\n");
			sb.append("if(showHex) {\n");
			sb.append("uint8_t segments[4] = {digits[(value >> 12) & 0x0F], digits[(value >> 8) & 0x0F], digits[(value >> 4) & 0x0F], digits[value & 0x0F]};\n");
			sb.append("segment_display_4x7_set_segments(segment_display, segments, 5, false);\n");
			sb.append("} else {\n");
			sb.append("uint8_t segments[4] = {digits[(value%10000)/1000], digits[(value%1000)/100], digits[(value%100)/10], digits[value%10]};\n");
			sb.append("segment_display_4x7_set_segments(segment_display, segments, 5, false);\n");
			sb.append("}\n");
			sb.append("}\n\n");
		}
		if(useLEDButton) {
			// for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
			for(String uid : ledButtons.keySet()) {
				sb.append("RGBLEDButton rlb_" + uid + ";\n");
			}
			for(String[] arr : ledButtonsLocalVariable.values()) {
				for(String var : arr) {
					sb.append("uint8_t " + var + " = 0;\n");
				}
			}
			// extractRGBforLEDButton(p);
			// }
		}
		if(useUS) {
			sb.append("uint16_t ultra_sonic_A;\n");
			sb.append("uint64_t us_A_last_cb_time = 0;\n");
			sb.append("char* uid_us_A = \"zpW\";\n");
			sb.append("void us_A_callback(uint16_t distance, void *data) {\n");
			sb.append("ultra_sonic_A = distance;\n");
			sb.append("us_A_last_cb_time = time_util_get_current_micros();\n");
			sb.append("}\n");
			sb.append("uint16_t ultra_sonic_B;\n");
			sb.append("uint64_t us_B_last_cb_time = 0;\n");
			sb.append("char* uid_us_B = \"zqN\";\n");
			sb.append("void us_B_callback(uint16_t distance, void *data) {\n");
			sb.append("ultra_sonic_B = distance;\n");
			sb.append("us_B_last_cb_time = time_util_get_current_micros();\n");
			sb.append("}\n\n");
		}
		if(useLaser) {
			sb.append("int16_t laser_distance;\n");
			sb.append("uint64_t laser_last_cb_time = 0;\n");
			sb.append("void laser_callback(uint16_t distance, void *data) {\n");
			sb.append("laser_distance = distance;\n");
			sb.append("laser_last_cb_time = time_util_get_current_micros();\n");
			sb.append("}\n\n");
		}
		if(useACC) {
			sb.append("float Q = 0.022;\n");
			sb.append("float R = 0.917;\n");
			sb.append("float estimates[9] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};\n");
			sb.append("float last_p[9] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};\n");
			sb.append("float kalman_filter(float value, int sensor) {\n");
			sb.append("if(sensor < 0 || sensor > 5) return -1;\n");
			sb.append("float temp_est = estimates[sensor];\n");
			sb.append("float p_temp = last_p[sensor] + Q;\n");
			sb.append("float K = p_temp * (1.0/(p_temp + R));\n");
			sb.append("float est = temp_est + K * (value - temp_est);\n");
			sb.append("float P = (1-K)*p_temp;\n");
			sb.append("last_p[sensor] = P;\n");
			sb.append("estimates[sensor] = est;\n");
			sb.append("return est;\n");
			sb.append("}\n\n");
			sb.append("float acceleration_X, acceleration_Y, acceleration_Z;\n");
			sb.append("uint64_t acc_last_cb_time = 0;\n");
			sb.append("void acc_callback(int16_t x, int16_t y, int16_t z, void* data) {\n");
			sb.append("if (abs(x - acceleration_X) > 2 * 10 * 100 ||\n");
			sb.append("abs(y - acceleration_Y) > 2 * 10 * 100 ||\n");
			sb.append("abs(z - acceleration_Z) > 2 * 10 * 100)\n");
			sb.append("    return;\n");
			sb.append("acceleration_X = kalman_filter(x, 0) / 100.0;\n");
			sb.append("acceleration_Y = kalman_filter(y, 1) / 100.0;\n");
			sb.append("acceleration_Z = kalman_filter(z, 2) / 100.0;\n");
			sb.append("acc_last_cb_time = time_util_get_current_micros();\n");
			sb.append("}\n\n");
			sb.append("float angularVelocity_X, angularVelocity_Y, angularVelocity_Z;\n");
			sb.append("uint64_t angV_last_cb_time = 0;\n");
			sb.append("void angV_callback(int16_t x, int16_t y, int16_t z, void* data) {\n");
			sb.append("angularVelocity_X = kalman_filter(x, 3) / 16;\n");
			sb.append("angularVelocity_Y = kalman_filter(y, 4) / 16;\n");
			sb.append("angularVelocity_Z = kalman_filter(z, 5) / 16;\n");
			sb.append("angV_last_cb_time = time_util_get_current_micros();\n");
			sb.append("}\n\n");
		}
		return sb.toString();
	}

	/**
	 * @param p
	 */
	private void extractRGBforLEDButton(Pair<PlatformConnectorUnit, Port> p) {
		final String UNKNOWN = "unknown";
		String uid = UNKNOWN;
		int rgb = -1;
		if(p.getFirst() instanceof LEDButtonRActuator) {
			LEDButtonRActuator sensor = (LEDButtonRActuator)p.getFirst();
			uid = sensor.getUniqueBrickletID();
			rgb = 0;
		}
		if(p.getFirst() instanceof LEDButtonGActuator) {
			LEDButtonGActuator sensor = (LEDButtonGActuator)p.getFirst();
			uid = sensor.getUniqueBrickletID();
			rgb = 1;
		}
		if(p.getFirst() instanceof LEDButtonBActuator) {
			LEDButtonBActuator sensor = (LEDButtonBActuator)p.getFirst();
			uid = sensor.getUniqueBrickletID();
			rgb = 2;
		}
		if(!uid.equals(UNKNOWN) && !ledButtons.containsKey(uid)) {
			String[] rgbArray = new String[3];
			rgbArray[rgb] = portName(p.getSecond());
			ledButtons.put(uid, rgbArray);
			String[] localVar = new String[3];
			localVar[rgb] =
					"rgbButton_" + uid + "_lastValue_" + (rgb == 0 ? "r" : rgb == 1 ? "g" : "b");
			ledButtonsLocalVariable.put(uid, localVar);
		} else if(ledButtons.containsKey(uid)) {
			String[] rgbArray = ledButtons.get(uid);
			rgbArray[rgb] = portName(p.getSecond());
			String[] localVar = ledButtonsLocalVariable.get(uid);
			localVar[rgb] =
					"rgbButton_" + uid + "_lastValue_" + (rgb == 0 ? "r" : rgb == 1 ? "g" : "b");
		}
	}

	/** Returns the port and ID identifier. */
	private String portName(Port port) {
		return port.getName().replace(' ', '_') + "_ID_" + port.getId();
	}

	/** Creates the read code for the given receiver and port. */
	private String createWriteCode(Transmitter transmitter, OutputPort outport) {
		if(transmitter instanceof ActuatorPWM) {
			ActuatorPWM act = (ActuatorPWM)transmitter;
			String result = "if(!noval_" + portName(outport) + ") {\n";
			if(act.getChannelID() == 0) {
				result +=
						"temp_actuator_device_set_target(maestro_fd, 0, (uint16_t)" +
								portName(outport) + ");\n";
			} else {
				result +=
						"temp_actuator_device_set_target(maestro_fd, 1, (uint16_t)" +
								portName(outport) + ");\n";
			}
			result += "}\n";
			return result;
		}
		if(transmitter instanceof ActuatorDigits) {
			ActuatorDigits digits = (ActuatorDigits)transmitter;
			String result = "if(!noval_" + portName(outport) + ") {\n";
			result +=
					"set_led_display(&segment_display, (uint16_t)" + portName(outport) + ", " +
							digits.isShowHexValue() + ");\n";
			result += "}\n";
			return result;
		}
		if(transmitter instanceof LEDButtonRActuator) {
			// TODO(JE): if a red actuator port is found the previously saved rgb values are used.
			LEDButtonRActuator digits = (LEDButtonRActuator)transmitter;
			String uniqueBrickletID = digits.getUniqueBrickletID();
			String[] rgb = ledButtons.get(uniqueBrickletID);
			String[] localRgbValue = ledButtonsLocalVariable.get(uniqueBrickletID);

			String result =
					"if(!noval_" + portName(outport) + " && (" + rgb[0] + " != " +
							localRgbValue[0] + " || " + rgb[1] + " != " + localRgbValue[1] +
							" || " + rgb[2] + " != " + localRgbValue[2] + ")) {\n";
			result +=
					"rgb_led_button_set_color(&rlb_" + uniqueBrickletID + ", " + rgb[0] + ", " +
							rgb[1] + ", " + rgb[2] + " );\n";
			result += localRgbValue[0] + " = " + rgb[0] + ";\n";
			result += localRgbValue[1] + " = " + rgb[1] + ";\n";
			result += localRgbValue[2] + " = " + rgb[2] + ";\n";
			result += "}\n";
			return result;
		}
		if(transmitter instanceof ConsoleOutput) {
			String result = "\tif(!noval_" + portName(outport) + ") {\n";
			result +=
					"\t\tprintf(\"" + outport.getName() + " = %i\\n\", " + portName(outport) +
							");\n";
			result += "\t} else {\n";
			result += "\t\tprintf(\"" + outport.getName() + " = NoVal\\n\");\n";
			result += "\t}\n";
			return result;
		}
		if(transmitter instanceof RumbleMagnitudeWeak) {
			useRumble = true;
			rumbleWeakPort = outport;
		}
		if(transmitter instanceof RumbleMagnitudeStrong) {
			useRumble = true;
			rumbleStrongPort = outport;
		}
		if(transmitter instanceof MotorControlOutput) {
			String result = "if(!noval_" + portName(outport) + ") {\n";
			result += "setVelocity(" + portName(outport) + ");\n";
			result += "}\n";
			return result;
		}
		return "";
	}

	/** Creates Code for Rumble */
	private String createRumbleCode() {
		StringBuilder sb = new StringBuilder();
		sb.append("rumblepad_set_rumble(");
		if(rumbleStrongPort != null) {
			sb.append("!noval_" + portName(rumbleStrongPort) + " ? " + portName(rumbleStrongPort) +
					" : 0, ");
		} else {
			sb.append("0, ");
		}
		if(rumbleWeakPort != null) {
			sb.append("!noval_" + portName(rumbleWeakPort) + " ? " + portName(rumbleWeakPort) +
					" : 0, ");
		} else {
			sb.append("0, ");
		}
		sb.append("cycle_time * 0.9, 0);\n");
		return sb.toString();
	}

	/** Creates the read code for the given receiver and port. */
	private String createReadCode(Receiver receiver, InputPort inport) {
		if(receiver instanceof CameraDistanceLeft) {
			StringBuilder sb = new StringBuilder();
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append(portName(inport)).append(" = camera_client_get_distance_left();\n");
			return sb.toString();
		}
		if(receiver instanceof CameraDistanceRight) {
			StringBuilder sb = new StringBuilder();
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append(portName(inport)).append(" = camera_client_get_distance_right();\n");
			return sb.toString();
		}
		if(receiver instanceof CameraDetectionStateLeft) {
			StringBuilder sb = new StringBuilder();
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append(portName(inport)).append(" = camera_client_get_detection_state_left();\n");
			return sb.toString();
		}
		if(receiver instanceof CameraDetectionStateRight) {
			StringBuilder sb = new StringBuilder();
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append(portName(inport)).append(" = camera_client_get_detection_state_right();\n");
			return sb.toString();
		}
		if(receiver instanceof CameraYawAngle) {
			StringBuilder sb = new StringBuilder();
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append(portName(inport)).append(" = camera_client_get_yaw_angle();\n");
			return sb.toString();
		}
		if(receiver instanceof UltraSonicSensor) {
			StringBuilder sb = new StringBuilder();
			UltraSonicSensor sensor = (UltraSonicSensor)receiver;
			if(nrOfUsSensors == 0) {
				sb.append("noval_" + portName(inport) + " = true;\n");
				sb.append("if (!strcmp(\"" + sensor.getUniqueBrickletID() + "\", uid_us_A)) {\n");
				sb.append("float diff_A = (curr_time - us_A_last_cb_time) / (1.0 * SECONDS_IN_MICROS);\n");
				sb.append("if (diff_A < 1.0) {\n");
				sb.append("noval_" + portName(inport) + " = false;\n");
				sb.append(portName(inport) + " = ultra_sonic_A;\n}\n");
				sb.append("}\n");
			} else if(nrOfUsSensors == 1) {
				sb.append("noval_" + portName(inport) + " = true;\n");
				sb.append("if (!strcmp(\"" + sensor.getUniqueBrickletID() + "\", uid_us_B)){\n");
				sb.append("float diff_B = (curr_time - us_B_last_cb_time) / (1.0 * SECONDS_IN_MICROS);\n");
				sb.append("if (diff_B < 1.0) {\n");
				sb.append("noval_" + portName(inport) + " = false;\n");
				sb.append(portName(inport) + " = ultra_sonic_B;\n}\n");
				sb.append("}\n");
			}
			nrOfUsSensors++;
			return sb.toString();
		}
		if(receiver instanceof LaserRangeSensor) {
			StringBuilder sb = new StringBuilder();
			sb.append("float diff_laser = (curr_time - laser_last_cb_time) / (1.0 * SECONDS_IN_MICROS);\n");
			sb.append("if (diff_laser < 1.0) {\n");
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append("if(laser_distance == 1){\n");
			sb.append(portName(inport) + " = 4000;\n");
			sb.append("} else {\n");
			sb.append(portName(inport) + " = laser_distance;\n");
			sb.append("}\n");
			sb.append("} else {\n");
			sb.append("noval_" + portName(inport) + " = true;\n");
			sb.append("}\n");
			return sb.toString();
		}
		if(receiver instanceof AccelerationXSensor) {
			StringBuilder sb = new StringBuilder();
			sb.append("float diff_accX = (curr_time - acc_last_cb_time) / (1.0 * SECONDS_IN_MICROS);\n");
			sb.append("if (diff_accX < 1.0) {\n");
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append(portName(inport) + " = acceleration_X;\n");
			sb.append("} else {\n");
			sb.append("noval_" + portName(inport) + " = true;\n");
			sb.append("}\n");
			return sb.toString();
		}
		if(receiver instanceof AccelerationYSensor) {
			StringBuilder sb = new StringBuilder();
			sb.append("float diff_accY = (curr_time - acc_last_cb_time) / (1.0 * SECONDS_IN_MICROS);\n");
			sb.append("if (diff_accY < 1.0) {\n");
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append(portName(inport) + " = acceleration_Y;\n");
			sb.append("} else {\n");
			sb.append("noval_" + portName(inport) + " = true;\n");
			sb.append("}\n");
			return sb.toString();
		}
		if(receiver instanceof AccelerationZSensor) {
			StringBuilder sb = new StringBuilder();
			sb.append("float diff_accZ = (curr_time - acc_last_cb_time) / (1.0 * SECONDS_IN_MICROS);\n");
			sb.append("if (diff_accZ < 1.0) {\n");
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append(portName(inport) + " = acceleration_Z;\n");
			sb.append("} else {\n");
			sb.append("noval_" + portName(inport) + " = true;\n");
			sb.append("}\n");
			return sb.toString();
		}
		if(receiver instanceof AngularVelocityXSensor) {
			StringBuilder sb = new StringBuilder();
			sb.append("float diff_angV_X = (curr_time - angV_last_cb_time) / (1.0 * SECONDS_IN_MICROS);\n");
			sb.append("if (diff_angV_X < 1.0) {\n");
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append(portName(inport) + " = angularVelocity_X;\n");
			sb.append("} else {\n");
			sb.append("noval_" + portName(inport) + " = true;\n");
			sb.append("}\n");
			return sb.toString();
		}
		if(receiver instanceof AngularVelocityYSensor) {
			StringBuilder sb = new StringBuilder();
			sb.append("float diff_angV_Y = (curr_time - angV_last_cb_time) / (1.0 * SECONDS_IN_MICROS);\n");
			sb.append("if (diff_angV_Y < 1.0) {\n");
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append(portName(inport) + " = angularVelocity_Y;\n");
			sb.append("} else {\n");
			sb.append("noval_" + portName(inport) + " = true;\n");
			sb.append("}\n");
			return sb.toString();
		}
		if(receiver instanceof AngularVelocityZSensor) {
			StringBuilder sb = new StringBuilder();
			sb.append("float diff_angV_Z = (curr_time - angV_last_cb_time) / (1.0 * SECONDS_IN_MICROS);\n");
			sb.append("if (diff_angV_Z < 1.0) {\n");
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append(portName(inport) + " = angularVelocity_Z;\n");
			sb.append("} else {\n");
			sb.append("noval_" + portName(inport) + " = true;\n");
			sb.append("}\n");
			return sb.toString();
		}
		if(receiver instanceof MotorControlInput) {
			StringBuilder sb = new StringBuilder();
			sb.append("noval_" + portName(inport) + " = false;\n");
			sb.append(portName(inport) + " = getVelocity();\n");
			return sb.toString();
		}
		if(receiver instanceof Button1) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_1", inport);
		}
		if(receiver instanceof Button2) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_2", inport);
		}
		if(receiver instanceof Button3) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_3", inport);
		}
		if(receiver instanceof Button4) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_4", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.gamepad.ButtonL1) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_L1", inport);
		}
		if(receiver instanceof ButtonL2) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_L2", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.gamepad.ButtonR1) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_R1", inport);
		}
		if(receiver instanceof ButtonR2) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_R2", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.gamepad.Right_StickX_Position) {
			return gamepadReadCode("gamepad_get_axis_position",
					"GAMEPAD_AXIS_RIGHT_STICK_HORIZONTAL", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.gamepad.Right_StickY_Position) {
			return gamepadReadCode("gamepad_get_axis_position",
					"GAMEPAD_AXIS_RIGHT_STICK_VERTICAL", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.gamepad.Left_StickX_Position) {
			return gamepadReadCode("gamepad_get_axis_position",
					"GAMEPAD_AXIS_LEFT_STICK_HORIZONTAL", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.gamepad.Left_StickY_Position) {
			return gamepadReadCode("gamepad_get_axis_position", "GAMEPAD_AXIS_LEFT_STICK_VERTICAL",
					inport);
		}
		if(receiver instanceof ButtonA) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_BUTTON_A", inport);
		}
		if(receiver instanceof ButtonB) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_BUTTON_B", inport);
		}
		if(receiver instanceof ButtonX) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_BUTTON_X", inport);
		}
		if(receiver instanceof ButtonY) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_BUTTON_Y", inport);
		}
		if(receiver instanceof ButtonStart) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_BUTTON_START", inport);
		}
		if(receiver instanceof ButtonSelect) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_BUTTON_SELECT", inport);
		}
		if(receiver instanceof ButtonHome) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_BUTTON_MODE", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.rumblepad.ButtonL1) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_BUTTON_L1", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.rumblepad.ButtonR1) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_BUTTON_R1", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.rumblepad.ButtonL3) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_BUTTON_L3", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.rumblepad.ButtonR3) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_BUTTON_R3", inport);
		}
		if(receiver instanceof L2_Position) {
			return gamepadReadCode("rumblepad_get_axis_position", "RUMBLEPAD_AXIS_L2", inport);
		}
		if(receiver instanceof R2_Position) {
			return gamepadReadCode("rumblepad_get_axis_position", "RUMBLEPAD_AXIS_R2", inport);
		}
		if(receiver instanceof DPadDown) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_DPAD_DOWN", inport);
		}
		if(receiver instanceof DPadUp) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_DPAD_UP", inport);
		}
		if(receiver instanceof DPadLeft) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_DPAD_LEFT", inport);
		}
		if(receiver instanceof DPadRight) {
			return gamepadReadCode("rumblepad_get_button_state", "RUMBLEPAD_DPAD_RIGHT", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.rumblepad.Left_StickX_Position) {
			return gamepadReadCode("rumblepad_get_axis_position",
					"RUMBLEPAD_AXIS_LEFT_STICK_HORIZONTAL", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.rumblepad.Left_StickY_Position) {
			return gamepadReadCode("rumblepad_get_axis_position",
					"RUMBLEPAD_AXIS_LEFT_STICK_VERTICAL", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.rumblepad.Right_StickX_Position) {
			return gamepadReadCode("rumblepad_get_axis_position",
					"RUMBLEPAD_AXIS_RIGHT_STICK_HORIZONTAL", inport);
		}
		if(receiver instanceof org.fortiss.af3.platform.raspberry.model.rumblepad.Right_StickY_Position) {
			return gamepadReadCode("rumblepad_get_axis_position",
					"RUMBLEPAD_AXIS_RIGHT_STICK_VERTICAL", inport);
		}
		return "";
	}

	/** Creates read code for gamepad state. */
	private String gamepadReadCode(String gamepadFun, String btnId, InputPort inport) {
		String noval = "noval_" + portName(inport) + " = false;\n";
		String setBtn = portName(inport) + " = " + gamepadFun + "(" + btnId + ");\n";
		return noval + setBtn;
	}

	/** Create the initialize code. */
	private String createInitCode(List<Pair<ExecutionUnit, Component>> deployedComponents) {
		int createdUsSensors = 0;
		StringBuilder sb = new StringBuilder();
		if(useCamera) {
			String addr = executionUnit.getCameraServerAddress();
			String port = executionUnit.getCameraServerPort();
			sb.append("camera_client_initialize(\"" + addr + "\", \"" + port + "\");\n");
		}
		if(useUS || useLaser || useDigits || useACC) {
			sb.append("IPConnection brick_connection;\n");
			sb.append("ipcon_create(&brick_connection);\n");
			sb.append("if(ipcon_connect(&brick_connection, BRICK_HOST, BRICK_PORT) < 0) {\n");
			sb.append("perror(\"Failed to connect to brick sub-system.\");\n");
			sb.append("return 1;\n");
			sb.append("}\n\n");
		}
		if(useUS) {
			System.out.println("useUs");
			for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
				if(createdUsSensors >= 2)
					break;
				if(p.getFirst() instanceof UltraSonicSensor) {
					UltraSonicSensor sensor = (UltraSonicSensor)p.getFirst();
					if(createdUsSensors == 0) {
						sb.append("DistanceUS DistanceUS_A;\n");
						sb.append("uid_us_A = \"" + sensor.getUniqueBrickletID() + "\";\n");
						sb.append("distance_us_create(&DistanceUS_A, \"" +
								sensor.getUniqueBrickletID() + "\", &brick_connection);\n");
						sb.append("distance_us_register_callback(&DistanceUS_A, DISTANCE_US_CALLBACK_DISTANCE, (void*)us_A_callback, NULL);\n");
						sb.append("distance_us_set_distance_callback_period(&DistanceUS_A, 10);\n");
					} else if(createdUsSensors == 1) {
						sb.append("DistanceUS DistanceUS_B;\n");
						sb.append("uid_us_B = \"" + sensor.getUniqueBrickletID() + "\";\n");
						sb.append("distance_us_create(&DistanceUS_B, \"" +
								sensor.getUniqueBrickletID() + "\", &brick_connection);\n");
						sb.append("distance_us_register_callback(&DistanceUS_B, DISTANCE_US_CALLBACK_DISTANCE, (void*)us_B_callback, NULL);\n");
						sb.append("distance_us_set_distance_callback_period(&DistanceUS_B, 10);\n");
					}
					createdUsSensors++;
				}
			}
		}
		if(useLaser) {
			for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
				if(p.getFirst() instanceof LaserRangeSensor) {
					LaserRangeSensor sensor = (LaserRangeSensor)p.getFirst();
					sb.append("LaserRangeFinder laser;\n");
					sb.append("laser_range_finder_create(&laser, \"" +
							sensor.getUniqueBrickletID() + "\", &brick_connection);\n");
					sb.append("laser_range_finder_enable_laser(&laser);\n");
					sb.append("laser_range_finder_register_callback(&laser, LASER_RANGE_FINDER_CALLBACK_DISTANCE, (void*)laser_callback, NULL);\n");
					sb.append("laser_range_finder_set_distance_callback_period(&laser, 10);\n");
					break;
				}
			}
		}
		if(useDigits) {
			for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
				if(p.getFirst() instanceof ActuatorDigits) {
					ActuatorDigits sensor = (ActuatorDigits)p.getFirst();
					sb.append("segment_display_4x7_create(&segment_display, \"" +
							sensor.getUniqueBrickletID() + "\", &brick_connection);\n");
					break;
				}
			}
		}
		if(useLEDButton) {
			for(String uid : ledButtons.keySet()) {
				sb.append("rgb_led_button_create(&rlb_" + uid + ", \"" + uid +
						"\", &brick_connection);\n");
			}
			// break; TODO(JE): those break statements only allow one element of each type!
		}
		if(useACC) {
			sb.append("IMUV2 imu;\n");
			for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
				if(p.getFirst() instanceof AccelerationXSensor ||
						p.getFirst() instanceof AccelerationYSensor ||
						p.getFirst() instanceof AccelerationZSensor ||
						p.getFirst() instanceof AngularVelocityXSensor ||
						p.getFirst() instanceof AngularVelocityYSensor ||
						p.getFirst() instanceof AngularVelocityZSensor) {
					UIDUnit sensor = (UIDUnit)p.getFirst();
					sb.append("imu_v2_create(&imu, \"" + sensor.getUniqueBrickletID() +
							"\", &brick_connection);\n");
					break;
				}
			}
			sb.append("imu_v2_register_callback(&imu, IMU_V2_CALLBACK_LINEAR_ACCELERATION, (void*)acc_callback, NULL);\n");
			sb.append("imu_v2_set_linear_acceleration_period(&imu, 5);\n");
			sb.append("imu_v2_register_callback(&imu, IMU_V2_CALLBACK_ANGULAR_VELOCITY, (void*)angV_callback, NULL);\n");
			sb.append("imu_v2_set_angular_velocity_period(&imu, 5);\n");
		}
		if(usePWM) {
			sb.append("maestro_fd = temp_actuator_initialize(\"/dev/Maestro0\");\n");
		}
		if(useGamepad) {
			sb.append("gamepad_configuration_t* gamepad_config = malloc(sizeof(gamepad_configuration_t));\n");
			sb.append("gamepad_config->device_id = \"/dev/input/js0\";\n");
			sb.append("gamepad_config->waiting_sleep_in_micros = " + WAITING_SLEEP_IN_MICROS +
					";\n");
			sb.append("gamepad_config->axis_callback = NULL;\n");
			sb.append("gamepad_config->button_callback = NULL;\n");
			sb.append("gamepad_initialize(gamepad_config);\n\n");
		}
		if(useRumblepad) {
			sb.append("while(access(\"/dev/input/event0\", F_OK) == -1) {\n");
			sb.append("sleep(1);\n");
			sb.append("printf(\"Could not access gamepad device. Trying again in 1s.\\n\");\n");
			sb.append("}\n");
			sb.append("rumblepad_configuration_t* rumblepad_config = malloc(sizeof(rumblepad_configuration_t));\n");
			sb.append("rumblepad_config->device_id = \"/dev/input/event0\";\n");
			sb.append("rumblepad_config->waiting_sleep_in_micros = " + WAITING_SLEEP_IN_MICROS +
					";\n");
			sb.append("rumblepad_config->axis_callback = NULL;\n");
			sb.append("rumblepad_config->button_callback = NULL;\n");
			sb.append("rumblepad_initialize(rumblepad_config);\n\n");
		}
		if(useVesc) {
			sb.append("initUSB(\"/dev/vesc\", B115200);\n");
		}
		for(Pair<ExecutionUnit, Component> p : deployedComponents) {
			Component c = p.getSecond();
			sb.append(makeCall("init", c));
		}
		return sb.toString();
	}

	/** Creates the function call code. */
	private String makeCall(String function, Component c) {
		StringBuilder sb = new StringBuilder();
		sb.append(function).append('_').append(c.getName());
		sb.append("_ID_").append(c.getId()).append("();\n");
		return sb.toString();
	}
}
