/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.executable;

import static org.fortiss.af3.component.utils.ComponentArchitectureUtils.isAtomicComponent;
import static org.fortiss.af3.expression.utils.ExpressionModelElementFactory.createDataDictionary;
import static org.fortiss.af3.generator.common.utils.CLanguageModelElementFactory.createCSourcePackage;
import static org.fortiss.af3.generator.common.utils.SourceModelElementFactory.createByteContentUnitForPluginFileInRCP;
import static org.fortiss.af3.platform.raspberry.generator.executable.HeaderCopyGenerator.copyHeaderFiles;
import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getConfigureFile;
import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getEclipseProjectFile;
import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getMakedefsFile;
import static org.fortiss.af3.platform.utils.PlatformArchitectureUtils.findAtomicTransmissionUnits;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getRootElement;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;
import static org.fortiss.tooling.kernel.utils.TransformationUtils.createTransformedObjectFor;

import java.util.ArrayList;
import java.util.List;

import org.conqat.lib.commons.collections.Pair;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.component.model.generator.ComponentProgram;
import org.fortiss.af3.deployment.generator.DeploymentExecutableTransformationContext;
import org.fortiss.af3.deployment.model.Deployment;
import org.fortiss.af3.expression.model.DataDictionary;
import org.fortiss.af3.expression.model.terms.IExpressionTerm;
import org.fortiss.af3.generator.common.model.c.AF3GeneratorCommonLanguagesCFactory;
import org.fortiss.af3.generator.common.model.c.CSourcePackage;
import org.fortiss.af3.platform.language.executable.ExecutionUnitExecutableBase;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.PlatformConnectorUnit;
import org.fortiss.af3.platform.model.TransmissionUnit;
import org.fortiss.af3.platform.raspberry.AF3PlatformRaspberryActivator;
import org.fortiss.af3.platform.raspberry.model.CanBus;
import org.fortiss.af3.platform.raspberry.model.RaspberryPi;
import org.fortiss.tooling.kernel.extension.data.ITransformationContext;
import org.fortiss.tooling.kernel.extension.exception.ChainTransformationFailedException;

/**
 * {@link ExecutionUnitExecutableBase} for the {@link RaspberryPi} platform.
 * 
 * @author eder
 * @author hoelzl
 * @author $Author: eder $
 * @version $Rev: 18709 $
 * @ConQAT.Rating RED Hash:
 */
public class RaspberryPIExecutable extends ExecutionUnitExecutableBase<RaspberryPi, CSourcePackage> {
	/** Folder name for library source files. */
	private static final String SRC_LIB_SUB_PACKAGE_NAME = "src-lib";
	/** Folder name for library include files. */
	public static final String INC_LIB_SUB_PACKAGE_NAME = "inc-lib";
	/** Folder name for linkable static library files. */
	private static final String LIB_SUB_PACKAGE_NAME = "lib";
	/** Default library names to be used always. */
	// Note that the order is important (GCC linker crap!)
	private static final String LIB_NAMES = "af3pihal rt pthread";
	/** Default library names extended with brick. */
	// Note that the order is important (GCC linker crap!)
	private static final String LIB_NAMES_WITH_BRICK = "brick " + LIB_NAMES;
	/** Default library names extended with vesc. */
	private static final String LIB_NAMES_WITH_VESC = LIB_NAMES_WITH_BRICK + " vesc";

	/** Constructor. */
	public RaspberryPIExecutable(RaspberryPi modelElement) {
		super(modelElement);
	}

	/** {@inheritDoc} */
	@Override
	protected CSourcePackage createExecutionUnitSourcePackage(String name,
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts, ITransformationContext context) {
		Deployment deployment = null;
		CanTransmissionCatalog catalog = null;
		if(context instanceof DeploymentExecutableTransformationContext) {
			DeploymentExecutableTransformationContext dContext =
					(DeploymentExecutableTransformationContext)context;
			deployment = dContext.getDeploymentExecutable().getDeployment();
			List<TransmissionUnit> tus =
					findAtomicTransmissionUnits(modelElement.getPlatformArchitecture());
			if(!tus.isEmpty() && tus.get(0) instanceof CanBus) {
				catalog =
						(CanTransmissionCatalog)dContext.getDeploymentExecutable()
								.getTransmissionCatalog(tus.get(0));
			}
		}
		// create the result package
		CSourcePackage sourcePackage = createCSourcePackage();
		sourcePackage.setBaseLocation(name);
		try {
			addDataDictionaryCode(context, sourcePackage);
			List<Component> atomics = new ArrayList<Component>();
			for(Pair<ExecutionUnit, Component> p : deployedComponents) {
				if(p.getFirst() != modelElement) {
					continue;
				}
				Component c = p.getSecond();
				if(isAtomicComponent(c)) {
					atomics.add(c);
				} else {
					atomics.addAll(c.findAtomicComponents());
				}
			}
			if(catalog != null) {
				addLogicalComponentCode(atomics, context, sourcePackage);
			} else {
				addLogicalComponentCodeSingleUnit(deployedComponents, context, sourcePackage);
			}
			addEclipseCProjectFiles(sourcePackage, modelElement.getName());
			addStaticLibrariesAndCode(sourcePackage);
			addConfigureAndMakedefsFiles(sourcePackage, modelElement.getName(), LIB_NAMES_WITH_VESC);
			addMainFile(deployment, catalog, sourcePackage, atomics, deployedComponents,
					deployedPorts, context);
		} catch(Exception ex) {
			error(AF3PlatformRaspberryActivator.getDefault(), ex.getMessage(), ex);
			ex.printStackTrace();
		}
		return sourcePackage;
	}

	/** Creates the main file by using the {@link MainGenerator} helper class. */
	private void addMainFile(Deployment deployment, CanTransmissionCatalog catalog,
			CSourcePackage sourcePackage, List<Component> atomics,
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts, ITransformationContext context) {
		CSourcePackage srcGenPack = (CSourcePackage)sourcePackage.getSrcGenPackage();
		if(catalog != null) {
			MultiUnitMainGenerator mg =
					new MultiUnitMainGenerator(deployment, catalog, modelElement, atomics,
							deployedComponents, deployedPorts, context);
			srcGenPack.addUnit(mg.createMultiUnitMain());
		} else {
			MainGenerator mg =
					new MainGenerator(modelElement, deployedComponents, deployedPorts, context);
			srcGenPack.addUnit(mg.createSingleUnitMain());
		}
	}

	/** Adds the build process files: configure and Makedefs. */
	private void addConfigureAndMakedefsFiles(CSourcePackage sourcePackage, String binaryName,
			String libNames) {
		sourcePackage.addUnit(getConfigureFile(binaryName, libNames));
		sourcePackage.addUnit(getMakedefsFile());
	}

	/** Adds the library code need by the deployment code generator. */
	private void addStaticLibrariesAndCode(CSourcePackage sourcePackage) throws Exception {
		CSourcePackage srcLib =
				AF3GeneratorCommonLanguagesCFactory.eINSTANCE.createCSourcePackage();
		srcLib.setBaseLocation(SRC_LIB_SUB_PACKAGE_NAME);
		sourcePackage.getSubPackages().add(srcLib);
		srcLib.addUnit(createByteContentUnitForPluginFileInRCP(
				AF3PlatformRaspberryActivator.PLUGIN_ID, "code-gen-hal/src", "data.c", false));

		CSourcePackage incLib =
				AF3GeneratorCommonLanguagesCFactory.eINSTANCE.createCSourcePackage();
		incLib.setBaseLocation(INC_LIB_SUB_PACKAGE_NAME);
		sourcePackage.getSubPackages().add(incLib);

		CSourcePackage lib = AF3GeneratorCommonLanguagesCFactory.eINSTANCE.createCSourcePackage();
		lib.setBaseLocation(LIB_SUB_PACKAGE_NAME);
		sourcePackage.getSubPackages().add(lib);
		lib.addUnit(createByteContentUnitForPluginFileInRCP(
				AF3PlatformRaspberryActivator.PLUGIN_ID, "code-gen-hal/lib", "libbrick.a", false));
		lib.addUnit(createByteContentUnitForPluginFileInRCP(
				AF3PlatformRaspberryActivator.PLUGIN_ID, "code-gen-hal/lib", "libaf3pihal.a", false));
		copyHeaderFiles(incLib);
	}

	/** Adds auxiliary files for Eclipse C project. */
	private void addEclipseCProjectFiles(CSourcePackage sourcePackage, String projectName) {
		sourcePackage.addUnit(getEclipseProjectFile(projectName));
	}

	/** Creates the code for all deployed components. */
	private void addLogicalComponentCode(List<Component> deployedAtomics,
			ITransformationContext context, CSourcePackage sourcePackage)
			throws ChainTransformationFailedException {
		for(Component cmp : deployedAtomics) {
			ComponentProgram cprog =
					createTransformedObjectFor(cmp, ComponentProgram.class, context);
			CSourcePackage cPack = createTransformedObjectFor(cprog, CSourcePackage.class, context);
			cPack.mergeInto(sourcePackage);
		}
	}

	/** Creates the code for all deployed components. */
	private void addLogicalComponentCodeSingleUnit(
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			ITransformationContext context, CSourcePackage sourcePackage)
			throws ChainTransformationFailedException {
		for(Pair<ExecutionUnit, Component> pair : deployedComponents) {
			if(pair.getFirst() != modelElement) {
				continue;
			}
			Component cmp = pair.getSecond();
			ComponentProgram cprog =
					createTransformedObjectFor(cmp, ComponentProgram.class, context);
			CSourcePackage cPack = createTransformedObjectFor(cprog, CSourcePackage.class, context);
			cPack.mergeInto(sourcePackage);
		}
	}

	/** Creates the code for the data dictionary. */
	private void
			addDataDictionaryCode(ITransformationContext context, CSourcePackage sourcePackage)
					throws ChainTransformationFailedException {
		DataDictionary dd = getRootElement(modelElement, DataDictionary.class);
		if(dd == null) {
			dd = createDataDictionary();
		}
		CSourcePackage ddPackage = createTransformedObjectFor(dd, CSourcePackage.class, context);
		ddPackage.mergeInto(sourcePackage);
	}

	/** {@inheritDoc} */
	@Override
	public IExpressionTerm getInitialization() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IExpressionTerm getTermination() {
		return null;
	}
}
