/*--------------------------------------------------------------------------+
$Id: RasPiCTemplates.java 22813 2018-01-12 15:25:51Z hoelzl $
|                                                                          |
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.templates;

import static org.fortiss.af3.generator.common.AF3GeneratorCommonActivator.getDefault;
import static org.fortiss.af3.generator.common.utils.SourceModelElementFactory.createStaticContentSourceUnit;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.io.IOException;
import java.io.InputStreamReader;

import org.antlr.stringtemplate.StringTemplate;
import org.antlr.stringtemplate.StringTemplateGroup;
import org.antlr.stringtemplate.language.DefaultTemplateLexer;
import org.fortiss.af3.generator.common.model.source.AbstractUnit;
import org.fortiss.af3.generator.common.model.source.StaticContentSourceUnit;

/**
 * Templates used by the RasPi C code generator.
 * 
 * @author hoelzl
 * @author $Author: hoelzl $
 * @version $Rev: 22813 $
 * @ConQAT.Rating RED Hash:
 */
public final class RasPiCTemplates {
	/** Returns the 'main.c' file configured using the given arguments. */
	public static AbstractUnit getMultiUnitMainCFile(String unitName, boolean coordinatorOrWorker,
			int canIdLocalUnit, String sensorVariables, String systemInitCode,
			String remoteUnitSetupCode, String systemIncludes, String componentPorts,
			String componentSetup, String componentImpl, int cycletimeInMillis,
			int threadSleepTimeInMicros) {
		StringTemplate template = makeTemplate("MultiUnitMainFile.stg", "MainFile");
		template.setAttribute("UNIT_NAME", unitName);
		template.setAttribute("COORDINATOR_OR_WORKER", coordinatorOrWorker ? "1" : "0");
		template.setAttribute("CAN_ID_LOCAL_UNIT", canIdLocalUnit);
		template.setAttribute("SENSOR_VARIABLES", sensorVariables);
		template.setAttribute("SYSTEM_INIT_CODE", systemInitCode);
		template.setAttribute("REMOTE_UNIT_SETUP_CODE", remoteUnitSetupCode);
		template.setAttribute("SYSTEM_INCLUDES", systemIncludes);
		template.setAttribute("COMPONENT_PORTS", componentPorts);
		template.setAttribute("COMPONENT_SETUP", componentSetup);
		template.setAttribute("COMPONENT_IMPL", componentImpl);
		template.setAttribute("CYCLE_TIME_IN_MILLIS", cycletimeInMillis);
		template.setAttribute("THREAD_SLEEPTIME_IN_MICROS", threadSleepTimeInMicros);
		StaticContentSourceUnit unit =
				createStaticContentSourceUnit("main.c", template.toString(), false);
		return unit;
	}

	/** Returns the 'main.c' file configured using the given arguments. */
	public static AbstractUnit
			getSingleUnitMainCFile(String unitName, int cycletimeInMillis, String systemIncludes,
					String systemInitCode, String workerCode, String sensorVariables) {
		StringTemplate template = makeTemplate("SingleUnitMainFile.stg", "MainFile");
		template.setAttribute("UNIT_NAME", unitName);
		template.setAttribute("CYCLE_TIME_IN_MILLIS", cycletimeInMillis);
		template.setAttribute("SYSTEM_INCLUDES", systemIncludes);
		template.setAttribute("SYSTEM_INIT_CODE", systemInitCode);
		template.setAttribute("WORKER_CODE", workerCode);
		template.setAttribute("SENSOR_VARIABLES", sensorVariables);
		StaticContentSourceUnit unit =
				createStaticContentSourceUnit("main.c", template.toString(), false);
		return unit;
	}

	/** Returns the '.project' file used for Eclipse project. */
	public static AbstractUnit getEclipseProjectFile(String projectName) {
		StringTemplate template = makeTemplate("ProjectFile.stg", "ProjectFile");
		template.setAttribute("PROJECT_NAME", projectName);
		StaticContentSourceUnit unit =
				createStaticContentSourceUnit(".project", template.toString(), false);
		return unit;
	}

	/** Returns the 'configure' file used to create a Makefile. */
	public static AbstractUnit getConfigureFile(String binaryName, String libNames) {
		StringTemplate template = makeTemplate("ConfigureFile.stg", "ConfigureFile");
		template.setAttribute("UNIT_NAME", binaryName);
		template.setAttribute("LIB_NAMES", libNames);
		return createStaticContentSourceUnit("configure", template.toString(), true);
	}

	/** Returns the 'Makedefs' file used to create a Makefile. */
	public static AbstractUnit getMakedefsFile() {
		StringTemplate template = makeTemplate("MakedefsFile.stg", "MakedefsFile");
		return createStaticContentSourceUnit("Makedefs", template.toString(), true);
	}

	/** Evaluates the given template. */
	protected static StringTemplate makeTemplate(String stgFile, String instance) {
		StringTemplateGroup stg = loadResourceSTG(stgFile);
		StringTemplate st = stg.getInstanceOf(instance);
		return st;
	}

	/** Loads the given template using the class loader of the given class. */
	private static StringTemplateGroup loadResourceSTG(String stgName) {
		InputStreamReader in =
				new InputStreamReader(RasPiCTemplates.class.getResourceAsStream(stgName));
		StringTemplateGroup templates = new StringTemplateGroup(in, DefaultTemplateLexer.class);
		try {
			in.close();
		} catch(IOException e) {
			error(getDefault(), "AF3 generator could not load stringtemplate " + stgName + ".", e);
		}
		return templates;
	}
}
